from datetime import datetime
from itertools import chain, groupby
from math import ceil
from typing import Any, Dict, Iterable, Iterator, List, Optional, Set, Union

import networkx as nx

from maggma.builders import Builder
from maggma.stores import Store
from maggma.utils import grouper

from emmet.builders.settings import EmmetBuildSettings
from emmet.core.utils import get_molecule_id, group_molecules, jsanitize, make_mol_graph
from emmet.core.qchem.molecule import (
    best_lot,
    evaluate_lot,
    evaluate_task_entry,
    MoleculeDoc,
)
from emmet.core.qchem.task import TaskDocument
from emmet.core.qchem.calc_types import LevelOfTheory, CalcType, TaskType


__author__ = "Evan Spotte-Smith <ewcspottesmith@lbl.gov>"


SETTINGS = EmmetBuildSettings()


def evaluate_molecule(
    mol_doc: MoleculeDoc,
    funct_scores: Dict[str, int] = SETTINGS.QCHEM_FUNCTIONAL_QUALITY_SCORES,
    basis_scores: Dict[str, int] = SETTINGS.QCHEM_BASIS_QUALITY_SCORES,
    solvent_scores: Dict[str, int] = SETTINGS.QCHEM_SOLVENT_MODEL_QUALITY_SCORES,
):
    """
    Helper function to order optimization calcs by
    - Level of theory
    - Electronic energy

    :param mol_doc: Molecule to be evaluated
    :param funct_scores: Scores for various density functionals
    :param basis_scores: Scores for various basis sets
    :param solvent_scores: Scores for various implicit solvent models
    :return:
    """

    opt_lot = None
    for origin in mol_doc.origins:
        if origin.name == "molecule":
            opt_lot = mol_doc.levels_of_theory[origin.task_id]
            if isinstance(opt_lot, LevelOfTheory):
                opt_lot = opt_lot.value

    if opt_lot is None:
        opt_eval = [0]
    else:
        opt_eval = evaluate_lot(opt_lot, funct_scores, basis_scores, solvent_scores)

    best = best_lot(mol_doc, funct_scores, basis_scores, solvent_scores)

    best_eval = evaluate_lot(best, funct_scores, basis_scores, solvent_scores)

    return (
        -1 * int(mol_doc.deprecated),
        sum(best_eval),
        sum(opt_eval),
        mol_doc.best_entries[best]["energy"],
    )


def _optimizing_solvent(mol_doc):
    """
    Returns which solvent was used to optimize this (associated) MoleculeDoc.

    Args:
        mol_doc: MoleculeDoc

    Returns:
        solvent (str)

    """

    for origin in mol_doc.origins:
        if origin.name.startswith("molecule"):
            solvent = mol_doc.solvents[origin.task_id]
            return solvent


class MoleculesAssociationBuilder(Builder):
    """
    The MoleculesAssociationBuilder matches Q-Chem task documents by composition
    and collects tasks associated with identical structures.
    The purpose of this builder is to group calculations in preparation for the
    MoleculesBuilder.

    The process is as follows:

        1.) Find all documents with the same formula
        2.) Select only task documents for the task_types we can select properties from
        3.) Aggregate task documents based on nuclear geometry
        4.) Create a MoleculeDoc from the group of task documents
    """

    def __init__(
        self,
        tasks: Store,
        assoc: Store,
        query: Optional[Dict] = None,
        settings: Optional[EmmetBuildSettings] = None,
        **kwargs,
    ):
        """
        Args:
            tasks:  Store of task documents
            assoc: Store of associated molecules documents to prepare
            query: dictionary to limit tasks to be analyzed
            settings: EmmetSettings to use in the build process
        """

        self.tasks = tasks
        self.assoc = assoc
        self.query = query if query else dict()
        self.settings = EmmetBuildSettings.autoload(settings)
        self.kwargs = kwargs

        super().__init__(sources=[tasks], targets=[assoc], **kwargs)

    def ensure_indexes(self):
        """
        Ensures indices on the collections needed for building
        """

        # Basic search index for tasks
        self.tasks.ensure_index("task_id")
        self.tasks.ensure_index("last_updated")
        self.tasks.ensure_index("state")
        self.tasks.ensure_index("formula_alphabetical")
        self.tasks.ensure_index("smiles")
        self.tasks.ensure_index("species_hash")

        # Search index for molecules
        self.assoc.ensure_index("molecule_id")
        self.assoc.ensure_index("last_updated")
        self.assoc.ensure_index("task_ids")
        self.assoc.ensure_index("formula_alphabetical")

    def prechunk(self, number_splits: int) -> Iterable[Dict]:  # pragma: no cover
        """Prechunk the molecule builder for distributed computation"""

        temp_query = dict(self.query)
        temp_query["state"] = "successful"

        self.logger.info("Finding tasks to process")
        all_tasks = list(self.tasks.query(temp_query, [self.tasks.key, "species_hash"]))

        processed_tasks = set(self.assoc.distinct("task_ids"))
        to_process_tasks = {d[self.tasks.key] for d in all_tasks} - processed_tasks
        to_process_hashes = {
            d["species_hash"]
            for d in all_tasks
            if d[self.tasks.key] in to_process_tasks
        }

        N = ceil(len(to_process_hashes) / number_splits)

        for hash_chunk in grouper(to_process_hashes, N):
            yield {"query": {"species_hash": {"$in": list(hash_chunk)}}}

    def get_items(self) -> Iterator[List[TaskDocument]]:
        """
        Gets all items to process into molecules (and other) documents.
        This does no datetime checking; relying on on whether
        task_ids are included in the molecules Store

        Returns:
            generator or list relevant tasks and molecules to process into documents
        """

        self.logger.info("Molecule association builder started")
        self.logger.info(
            f"Allowed task types: {[task_type.value for task_type in self.settings.QCHEM_ALLOWED_TASK_TYPES]}"
        )

        self.logger.info("Setting indexes")
        self.ensure_indexes()

        # Save timestamp to mark buildtime
        self.timestamp = datetime.utcnow()

        # Get all processed tasks
        temp_query = dict(self.query)
        temp_query["state"] = "successful"

        self.logger.info("Finding tasks to process")
        all_tasks = list(self.tasks.query(temp_query, [self.tasks.key, "species_hash"]))

        processed_tasks = set(self.assoc.distinct("task_ids"))
        to_process_tasks = {d[self.tasks.key] for d in all_tasks} - processed_tasks
        to_process_hashes = {
            d["species_hash"]
            for d in all_tasks
            if d[self.tasks.key] in to_process_tasks
        }

        self.logger.info(f"Found {len(to_process_tasks)} unprocessed tasks")
        self.logger.info(f"Found {len(to_process_hashes)} unprocessed hashes")

        # Set total for builder bars to have a total
        self.total = len(to_process_hashes)

        projected_fields = [
            "last_updated",
            "task_id",
            "formula_alphabetical",
            "species_hash",
            "coord_hash",
            "smiles",
            "orig",
            "tags",
            "walltime",
            "cputime",
            "output",
            "calcs_reversed",
            "special_run_type",
            "custom_smd",
            "critic2",
        ]

        for shash in to_process_hashes:
            tasks_query = dict(temp_query)
            tasks_query["species_hash"] = shash
            tasks = list(
                self.tasks.query(criteria=tasks_query, properties=projected_fields)
            )
            to_yield = list()
            for t in tasks:
                # TODO: Validation
                # basic validation here ensures that tasks with invalid levels of
                # theory don't halt the build pipeline
                try:
                    task = TaskDocument(**t)
                    to_yield.append(task)
                except Exception as e:
                    self.logger.info(
                        f"Processing task {t['task_id']} failed with Exception - {e}"
                    )
                    continue

            yield to_yield

    def process_item(self, tasks: List[TaskDocument]) -> List[Dict]:
        """
        Process the tasks into a MoleculeDoc

        Args:
            tasks [TaskDocument] : a list of task docs

        Returns:
            [dict] : a list of new molecule docs
        """

        if len(tasks) == 0:
            return list()
        shash = tasks[0].species_hash
        task_ids = [task.task_id for task in tasks]
        self.logger.debug(f"Processing {shash} : {task_ids}")
        molecules = list()

        for group in self.filter_and_group_tasks(tasks):
            try:
                doc = MoleculeDoc.from_tasks(group)
                molecules.append(doc)
            except Exception as e:
                failed_ids = list({t_.task_id for t_ in group})
                doc = MoleculeDoc.construct_deprecated_molecule(group)
                doc.warnings.append(str(e))
                molecules.append(doc)
                self.logger.warning(
                    f"Failed making molecule for {failed_ids}."
                    f" Inserted as deprecated molecule: {doc.molecule_id}"
                )

        self.logger.debug(f"Produced {len(molecules)} molecules for {shash}")

        return jsanitize([mol.model_dump() for mol in molecules], allow_bson=True)

    def update_targets(self, items: List[List[Dict]]):
        """
        Inserts the new molecules into the molecules collection

        Args:
            items [[dict]]: A list of molecules to update
        """

        docs = list(chain.from_iterable(items))  # type: ignore

        for item in docs:
            item.update({"_bt": self.timestamp})

        molecule_ids = list({item["molecule_id"] for item in docs})

        if len(items) > 0:
            self.logger.info(f"Updating {len(docs)} molecules")
            self.assoc.remove_docs({self.assoc.key: {"$in": molecule_ids}})
            self.assoc.update(
                docs=docs,
                key=["molecule_id"],
            )
        else:
            self.logger.info("No items to update")

    def filter_and_group_tasks(
        self, tasks: List[TaskDocument]
    ) -> Iterator[List[TaskDocument]]:
        """
        Groups tasks by identical structure
        """

        filtered_tasks = [
            task
            for task in tasks
            if any(
                allowed_type is task.task_type
                for allowed_type in self.settings.QCHEM_ALLOWED_TASK_TYPES
            )
        ]

        molecules = list()

        for idx, task in enumerate(filtered_tasks):
            if task.output.optimized_molecule:
                m = task.output.optimized_molecule
            else:
                m = task.output.initial_molecule
            m.ind: int = idx  # type: ignore
            molecules.append(m)

        grouped_molecules = group_molecules(molecules)
        for group in grouped_molecules:
            grouped_tasks = [filtered_tasks[mol.ind] for mol in group]  # type: ignore
            yield grouped_tasks


class MoleculesBuilder(Builder):
    """
    The MoleculesBuilder collects MoleculeDocs from the MoleculesAssociationBuilder
    and groups them by key properties (charge, spin multiplicity, bonding).
    Then, the best molecular structure is identified (based on electronic energy),
    and this document becomes the representative MoleculeDoc.

    The process is as follows:

        1.) Find all documents with the same formula
        2.) Group documents based on charge, spin, and bonding
        3.) Create a MoleculeDoc from the group of task documents
    """

    def __init__(
        self,
        assoc: Store,
        molecules: Store,
        query: Optional[Dict] = None,
        settings: Optional[EmmetBuildSettings] = None,
        **kwargs,
    ):
        """
        Args:
            assoc:  Store of associated molecules documents, created by MoleculesAssociationBuilder
            molecules: Store of processed molecules documents
            query: dictionary to limit tasks to be analyzed
            settings: EmmetSettings to use in the build process
        """

        self.assoc = assoc
        self.molecules = molecules
        self.query = query if query else dict()
        self.settings = EmmetBuildSettings.autoload(settings)
        self.kwargs = kwargs

        super().__init__(sources=[assoc], targets=[molecules], **kwargs)

    def ensure_indexes(self):
        """
        Ensures indices on the collections needed for building
        """

        # Search index for associated molecules
        self.assoc.ensure_index("molecule_id")
        self.assoc.ensure_index("last_updated")
        self.assoc.ensure_index("task_ids")
        self.assoc.ensure_index("formula_alphabetical")

        # Search index for molecules
        self.molecules.ensure_index("molecule_id")
        self.molecules.ensure_index("last_updated")
        self.molecules.ensure_index("task_ids")
        self.molecules.ensure_index("formula_alphabetical")

    def prechunk(self, number_splits: int) -> Iterable[Dict]:  # pragma: no cover
        """Prechunk the molecule builder for distributed computation"""

        temp_query = dict(self.query)
        temp_query["deprecated"] = False

        self.logger.info("Finding documents to process")
        all_assoc = list(
            self.assoc.query(
                temp_query,
                [
                    self.assoc.key,
                    "formula_alphabetical",
                    "species_hash",
                    "charge",
                    "spin_multiplicity",
                ],
            )
        )

        # Should be using species hash, rather than coord hash, at this point
        processed_docs = set(list(self.molecules.distinct("molecule_id")))
        assoc_ids = set()

        xyz_species_id_map = dict()
        for d in all_assoc:
            this_id = "{}-{}-{}-{}".format(
                d["species_hash"],
                d["formula_alphabetical"].replace(" ", ""),
                str(int(d["charge"])).replace("-", "m"),
                str(int(d["spin_multiplicity"])),
            )
            assoc_ids.add(this_id)
            xyz_species_id_map[d[self.assoc.key]] = this_id
        to_process_docs = assoc_ids - processed_docs

        to_process_forms = {
            d["formula_alphabetical"]
            for d in all_assoc
            if xyz_species_id_map[d[self.assoc.key]] in to_process_docs
        }

        N = ceil(len(to_process_forms) / number_splits)

        for formula_chunk in grouper(to_process_forms, N):
            yield {"query": {"formula_alphabetical": {"$in": list(formula_chunk)}}}

    def get_items(self) -> Iterator[List[Dict]]:
        """
        Gets all items to process into molecules (and other) documents.
        This does no datetime checking; relying on on whether
        task_ids are included in the molecules Store

        Returns:
            generator or list relevant tasks and molecules to process into documents
        """

        self.logger.info("Molecules builder started")
        self.logger.info("Setting indexes")
        self.ensure_indexes()

        # Save timestamp to mark buildtime
        self.timestamp = datetime.utcnow()

        # Get all processed molecules
        temp_query = dict(self.query)
        temp_query["deprecated"] = False

        self.logger.info("Finding documents to process")
        all_assoc = list(
            self.assoc.query(
                temp_query,
                [
                    self.assoc.key,
                    "formula_alphabetical",
                    "species_hash",
                    "charge",
                    "spin_multiplicity",
                ],
            )
        )

        # Should be using species hash, rather than coord hash, at this point
        processed_docs = set(list(self.molecules.distinct("molecule_id")))
        assoc_ids = set()

        xyz_species_id_map = dict()
        for d in all_assoc:
            this_id = "{}-{}-{}-{}".format(
                d["species_hash"],
                d["formula_alphabetical"].replace(" ", ""),
                str(int(d["charge"])).replace("-", "m"),
                str(int(d["spin_multiplicity"])),
            )
            assoc_ids.add(this_id)
            xyz_species_id_map[d[self.assoc.key]] = this_id
        to_process_docs = assoc_ids - processed_docs

        to_process_forms = {
            d["formula_alphabetical"]
            for d in all_assoc
            if xyz_species_id_map[d[self.assoc.key]] in to_process_docs
        }

        self.logger.info(f"Found {len(to_process_docs)} unprocessed documents")
        self.logger.info(f"Found {len(to_process_forms)} unprocessed formulas")

        # Set total for builder bars to have a total
        self.total = len(to_process_forms)

        for formula in to_process_forms:
            assoc_query = dict(temp_query)
            assoc_query["formula_alphabetical"] = formula
            assoc = list(self.assoc.query(criteria=assoc_query))

            yield assoc

    def process_item(self, items: List[Dict]) -> List[Dict]:
        """
        Process the tasks into a MoleculeDoc

        Args:
            tasks List[Dict] : a list of task docs

        Returns:
            [dict] : a list of new molecule docs
        """

        assoc = [MoleculeDoc(**item) for item in items]
        formula = assoc[0].formula_alphabetical
        mol_ids = [a.molecule_id for a in assoc]
        self.logger.debug(f"Processing {formula} : {mol_ids}")

        complete_mol_docs = list()

        # This is only slightly unholy
        # Need to combine many variables of the various constituent associated docs
        # into one MoleculeDoc, where the best associated doc for each solvent is taken
        for group in self.group_mol_docs(assoc):
            # Maybe all are disconnected and therefore none get grouped?
            if len(group) == 0:
                continue

            docs_by_solvent = dict()
            mols_by_solvent = dict()
            mol_lots = dict()

            task_ids = list()
            calc_types = dict()
            task_types = dict()
            levels_of_theory = dict()
            solvents = dict()
            lot_solvents = dict()
            unique_calc_types: Set[Union[str, CalcType]] = set()
            unique_task_types: Set[Union[str, TaskType]] = set()
            unique_levels_of_theory: Set[Union[str, LevelOfTheory]] = set()
            unique_solvents: Set[str] = set()
            unique_lot_solvents: Set[str] = set()
            origins = list()
            entries = list()
            best_entries: Dict[str, Any] = dict()
            constituent_molecules = list()
            similar_molecules = list()

            base_doc: Optional[MoleculeDoc] = None

            # Grab best doc for each solvent
            # A doc is given a solvent based on how the molecule was optimized
            for solv, subgroup in groupby(
                sorted(group, key=_optimizing_solvent), key=_optimizing_solvent
            ):
                sorted_docs = sorted(subgroup, key=evaluate_molecule)
                docs_by_solvent[solv] = sorted_docs[0]
                mols_by_solvent[solv] = sorted_docs[0].molecule
                mol_lots[solv] = sorted_docs[0].levels_of_theory[
                    sorted_docs[0].origins[0].task_id
                ]
                constituent_molecules.append(sorted_docs[0].molecule_id)

                if len(sorted_docs) > 1:
                    for m in sorted_docs[1:]:
                        if m.molecule_id not in constituent_molecules:
                            similar_molecules.append(m.molecule_id)

                if base_doc is None:
                    base_doc = docs_by_solvent[solv]

            if base_doc is None:
                continue

            else:
                # Compile data on each constituent doc
                for solv, doc in docs_by_solvent.items():
                    task_ids.extend(doc.task_ids)
                    calc_types.update(doc.calc_types)
                    task_types.update(doc.task_types)
                    levels_of_theory.update(doc.levels_of_theory)
                    solvents.update(doc.solvents)
                    lot_solvents.update(doc.lot_solvents)
                    unique_calc_types = unique_calc_types.union(
                        set(doc.unique_calc_types)
                    )
                    unique_task_types = unique_task_types.union(
                        set(doc.unique_task_types)
                    )
                    unique_levels_of_theory = unique_levels_of_theory.union(
                        set(doc.unique_levels_of_theory)
                    )
                    unique_solvents = unique_solvents.union(set(doc.unique_solvents))
                    unique_lot_solvents = unique_lot_solvents.union(
                        set(doc.unique_lot_solvents)
                    )
                    origins.extend(doc.origins)
                    entries.extend(doc.entries)

                    for lot_solv, entry in doc.best_entries.items():
                        if lot_solv in best_entries:
                            current_eval = evaluate_task_entry(best_entries[lot_solv])
                            this_eval = evaluate_task_entry(entry)
                            if this_eval < current_eval:
                                best_entries[lot_solv] = entry
                        else:
                            best_entries[lot_solv] = entry

                # Assign new doc info
                base_doc.molecule_id = get_molecule_id(
                    base_doc.molecule, node_attr="specie"
                )
                base_doc.molecules = mols_by_solvent
                base_doc.molecule_levels_of_theory = mol_lots
                base_doc.task_ids = task_ids
                base_doc.calc_types = calc_types
                base_doc.task_types = task_types
                base_doc.levels_of_theory = levels_of_theory
                base_doc.solvents = solvents
                base_doc.lot_solvents = lot_solvents
                base_doc.unique_calc_types = unique_calc_types
                base_doc.unique_task_types = unique_task_types
                base_doc.unique_levels_of_theory = unique_levels_of_theory
                base_doc.unique_solvents = unique_solvents
                base_doc.unique_lot_solvents = unique_lot_solvents
                base_doc.origins = origins
                base_doc.entries = entries
                base_doc.best_entries = best_entries
                base_doc.constituent_molecules = constituent_molecules
                base_doc.similar_molecules = similar_molecules

                complete_mol_docs.append(base_doc)

        self.logger.debug(f"Produced {len(complete_mol_docs)} molecules for {formula}")

        return jsanitize(
            [mol.model_dump() for mol in complete_mol_docs], allow_bson=True
        )

    def update_targets(self, items: List[List[Dict]]):
        """
        Inserts the new molecules into the molecules collection

        Args:
            items [[dict]]: A list of molecules to update
        """

        self.logger.debug(f"Updating {len(items)} molecules")

        docs = list(chain.from_iterable(items))  # type: ignore

        # Add timestamp, add prefix to molecule id
        for item in docs:
            molid = item["molecule_id"]

            item.update({"_bt": self.timestamp})

            for entry in item["entries"]:
                entry["entry_id"] = molid

        molecule_ids = list({item["molecule_id"] for item in docs})

        if len(items) > 0:
            self.logger.info(f"Updating {len(docs)} molecules")
            self.molecules.remove_docs({self.molecules.key: {"$in": molecule_ids}})
            self.molecules.update(
                docs=docs,
                key=["molecule_id"],
            )
        else:
            self.logger.info("No items to update")

    def group_mol_docs(self, assoc: List[MoleculeDoc]) -> Iterator[List[MoleculeDoc]]:
        """
        Groups molecules by:
            - highest level of theory
            - charge
            - spin multiplicity
            - bonding (molecule graph isomorphism)
            - solvent environment used for the structure
        """

        # Molecules are already grouped by formula

        # First, group by charge, spin multiplicity
        # Then group by graph isomorphism, using OpenBabelNN + metal_edge_extender

        def charge_spin(mol_doc):
            return (mol_doc.charge, mol_doc.spin_multiplicity)

        # Group by charge and spin
        for c_s, group in groupby(sorted(assoc, key=charge_spin), key=charge_spin):
            subgroups: List[Dict[str, Any]] = list()
            for mol_doc in group:
                mol_graph = make_mol_graph(mol_doc.molecule)
                mol_hash = mol_doc.species_hash

                # Finally, group by graph isomorphism
                # When bonding is defined by OpenBabelNN + metal_edge_extender
                # Unconnected molecule graphs are discarded at this step
                # TODO: What about molecules that would be connected under a different
                # TODO: bonding scheme? For now, ¯\_(ツ)_/¯
                # TODO: MAKE ClusterBuilder FOR THIS PURPOSE
                if nx.is_connected(mol_graph.graph.to_undirected()):
                    matched = False

                    for subgroup in subgroups:
                        if mol_hash == subgroup["hash"]:
                            subgroup["mol_docs"].append(mol_doc)
                            matched = True
                            break

                    if not matched:
                        subgroups.append({"hash": mol_hash, "mol_docs": [mol_doc]})

            self.logger.debug(f"Unique hashes: {[x['hash'] for x in subgroups]}")

            for subgroup in subgroups:
                yield subgroup["mol_docs"]



import pytest
from maggma.stores import JSONStore, MemoryStore

from emmet.builders.qchem.molecules import MoleculesAssociationBuilder, MoleculesBuilder
from emmet.builders.molecules.thermo import ThermoBuilder


__author__ = "Evan Spotte-Smith <ewcspottesmith@lbl.gov>"


@pytest.fixture(scope="session")
def tasks_store(test_dir):
    return JSONStore(test_dir / "C2H4.json.gz")


@pytest.fixture(scope="session")
def mol_store(tasks_store):
    assoc_store = MemoryStore(key="molecule_id")
    stage_one = MoleculesAssociationBuilder(tasks=tasks_store, assoc=assoc_store)
    stage_one.run()

    mol_store = MemoryStore(key="molecule_id")
    stage_two = MoleculesBuilder(assoc=assoc_store, molecules=mol_store)
    stage_two.run()

    return mol_store


@pytest.fixture(scope="session")
def thermo_store():
    return MemoryStore()


def test_thermo_builder(tasks_store, mol_store, thermo_store):
    builder = ThermoBuilder(tasks_store, mol_store, thermo_store)
    builder.run()

    assert thermo_store.count() == 20
    assert thermo_store.count({"vibrational_enthalpy": None}) == 0



from math import ceil

from maggma.builders import Builder
from maggma.utils import grouper

from emmet.core.mpid import MPID
from emmet.core.summary import SummaryDoc, HasProps
from emmet.core.utils import jsanitize
from emmet.core.thermo import ThermoType


class SummaryBuilder(Builder):
    def __init__(
        self,
        materials,
        thermo,
        xas,
        chemenv,
        absorption,
        grain_boundaries,
        electronic_structure,
        magnetism,
        elasticity,
        dielectric,
        piezoelectric,
        phonon,
        insertion_electrodes,
        substrates,
        surfaces,
        oxi_states,
        eos,
        provenance,
        charge_density_index,
        summary,
        thermo_type=ThermoType.GGA_GGA_U_R2SCAN.value,
        chunk_size=100,
        query=None,
        **kwargs,
    ):
        self.materials = materials
        self.thermo = thermo
        self.xas = xas
        self.chemenv = chemenv
        self.absorption = absorption
        self.grain_boundaries = grain_boundaries
        self.electronic_structure = electronic_structure
        self.magnetism = magnetism
        self.elasticity = elasticity
        self.dielectric = dielectric
        self.piezoelectric = piezoelectric
        self.phonon = phonon
        self.insertion_electrodes = insertion_electrodes
        self.substrates = substrates
        self.surfaces = surfaces
        self.oxi_states = oxi_states
        self.eos = eos
        self.provenance = provenance
        self.charge_density_index = charge_density_index

        self.thermo_type = thermo_type

        self.summary = summary
        self.chunk_size = chunk_size
        self.query = query if query else {}

        super().__init__(
            sources=[
                materials,
                thermo,
                xas,
                chemenv,
                absorption,
                grain_boundaries,
                electronic_structure,
                magnetism,
                elasticity,
                dielectric,
                piezoelectric,
                phonon,
                insertion_electrodes,
                surfaces,
                oxi_states,
                substrates,
                eos,
                provenance,
                charge_density_index,
            ],
            targets=[summary],
            chunk_size=chunk_size,
            **kwargs,
        )

    def get_items(self):
        """
        Gets all items to process

        Returns:
            list of relevant materials and data
        """

        self.logger.info("Summary Builder Started")

        q = dict(self.query)

        mat_ids = self.materials.distinct(field=self.materials.key, criteria=q)
        summary_ids = self.summary.distinct(field=self.summary.key, criteria=q)

        summary_set = set(mat_ids) - set(summary_ids)

        self.total = len(summary_set)

        self.logger.debug("Processing {} materials.".format(self.total))

        for entry in summary_set:
            materials_doc = self.materials.query_one({self.materials.key: entry})

            valid_static_tasks = set(
                [
                    task_id
                    for task_id, task_type in materials_doc["task_types"].items()
                    if task_type == "Static"
                ]
            ) - set(materials_doc["deprecated_tasks"])

            all_tasks = list(materials_doc["task_types"].keys())

            data = {
                HasProps.materials.value: materials_doc,
                HasProps.thermo.value: self.thermo.query_one(
                    {self.materials.key: entry, "thermo_type": str(self.thermo_type)}
                ),
                HasProps.xas.value: list(
                    self.xas.query({self.xas.key: {"$in": all_tasks}})
                ),
                HasProps.grain_boundaries.value: list(
                    self.grain_boundaries.query({self.grain_boundaries.key: entry})
                ),
                HasProps.electronic_structure.value: self.electronic_structure.query_one(
                    {self.electronic_structure.key: entry}
                ),
                HasProps.magnetism.value: self.magnetism.query_one(
                    {self.magnetism.key: entry}
                ),
                HasProps.elasticity.value: self.elasticity.query_one(
                    {self.elasticity.key: {"$in": all_tasks}}
                ),
                HasProps.dielectric.value: self.dielectric.query_one(
                    {self.dielectric.key: entry}
                ),
                HasProps.piezoelectric.value: self.piezoelectric.query_one(
                    {self.piezoelectric.key: entry}
                ),
                HasProps.phonon.value: self.phonon.query_one(
                    {self.phonon.key: {"$in": all_tasks}},
                    [self.phonon.key],
                ),
                HasProps.insertion_electrodes.value: list(
                    self.insertion_electrodes.query(
                        {"material_ids": entry},
                        [self.insertion_electrodes.key],
                    )
                ),
                HasProps.surface_properties.value: self.surfaces.query_one(
                    {self.surfaces.key: {"$in": all_tasks}}
                ),
                HasProps.substrates.value: list(
                    self.substrates.query(
                        {self.substrates.key: {"$in": all_tasks}}, [self.substrates.key]
                    )
                ),
                HasProps.oxi_states.value: self.oxi_states.query_one(
                    {self.oxi_states.key: entry}
                ),
                HasProps.eos.value: self.eos.query_one(
                    {self.eos.key: {"$in": all_tasks}}, [self.eos.key]
                ),
                HasProps.chemenv.value: self.chemenv.query_one(
                    {self.chemenv.key: entry}
                ),
                HasProps.absorption.value: self.absorption.query_one(
                    {self.absorption.key: entry}
                ),
                HasProps.provenance.value: self.provenance.query_one(
                    {self.provenance.key: entry}
                ),
                HasProps.charge_density.value: self.charge_density_index.query_one(
                    {"task_id": {"$in": list(valid_static_tasks)}}, ["task_id"]
                ),
            }

            sub_fields = {}

            for collection, sub_field in sub_fields.items():
                if data[collection] is not None:
                    data[collection] = (
                        data[collection][sub_field]
                        if (sub_field in data[collection])
                        and (data[collection][sub_field] != {})
                        else None
                    )

            yield data

    def prechunk(self, number_splits: int):  # pragma: no cover
        """
        Prechunk method to perform chunking by the key field
        """
        q = dict(self.query)

        keys = self.summary.newer_in(self.materials, criteria=q, exhaustive=True)

        N = ceil(len(keys) / number_splits)
        for split in grouper(keys, N):
            yield {"query": {self.materials.key: {"$in": list(split)}}}

    def process_item(self, item):
        material_id = MPID(item[HasProps.materials.value]["material_id"])
        doc = SummaryDoc.from_docs(material_id=material_id, **item)
        return jsanitize(doc.model_dump(exclude_none=False), allow_bson=True)

    def update_targets(self, items):
        """
        Copy each summary doc to the store

        Args:
            items ([dict]): A list of dictionaries of mpid document pairs to update
        """
        items = list(filter(None, items))

        if len(items) > 0:
            self.logger.info("Inserting {} summary docs".format(len(items)))
            self.summary.update(docs=items)
        else:
            self.logger.info("No summary entries to update")



from collections import defaultdict
import copy
from datetime import datetime
from itertools import chain, groupby
from math import ceil
from typing import Any, Dict, Iterable, Iterator, List, Optional, Union

from maggma.builders import Builder
from maggma.core import Store
from maggma.utils import grouper

from emmet.core.qchem.task import TaskDocument
from emmet.core.qchem.molecule import MoleculeDoc
from emmet.core.molecules.bonds import metals
from emmet.core.molecules.thermo import MoleculeThermoDoc
from emmet.core.molecules.redox import RedoxDoc
from emmet.core.utils import confirm_molecule, get_graph_hash, jsanitize
from emmet.builders.settings import EmmetBuildSettings


__author__ = "Evan Spotte-Smith"

SETTINGS = EmmetBuildSettings()


class RedoxBuilder(Builder):
    """
    The RedoxBuilder extracts the highest-quality redox data (vertical and
    adiabatic reduction and oxidation potentials, etc.)
    from a MoleculeDoc (lowest electronic energy, highest level of theory).

    The process is as follows:
        1. Gather MoleculeDocs by formula
        2. Further group based on (covalent) isomorphism and charge
        3. For each MoleculeDoc:
            3a. Identify relevant MoleculeThermoDocs
            3b. Look for single-point energy calculations conducted at the
            molecule's charge +- 1. These will be used to calculation
            vertical electron affinities and ionization energies
            3c. Group MoleculeThermoDocs and single-point calculations based on solvent
            and level of theory
        4. Construct RedoxDocs by looking for molecules (with associated
            calculations) that:
            - Have charges that differ by +- 1
            - Use the same solvent and level of theory
    """

    def __init__(
        self,
        tasks: Store,
        molecules: Store,
        thermo: Store,
        redox: Store,
        query: Optional[Dict] = None,
        settings: Optional[EmmetBuildSettings] = None,
        **kwargs,
    ):
        self.tasks = tasks
        self.molecules = molecules
        self.thermo = thermo
        self.redox = redox
        self.query = query if query else dict()
        self.settings = EmmetBuildSettings.autoload(settings)
        self.kwargs = kwargs

        super().__init__(sources=[tasks, molecules, thermo], targets=[redox], **kwargs)
        # Uncomment in case of issue with mrun not connecting automatically to collections
        # for i in [self.tasks, self.molecules, self.thermo, self.redox]:
        #     try:
        #         i.connect()
        #     except Exception as e:
        #         print("Could not connect,", e)

    def ensure_indexes(self):
        """
        Ensures indices on the collections needed for building
        """

        # Basic search index for tasks
        self.tasks.ensure_index("task_id")
        self.tasks.ensure_index("last_updated")
        self.tasks.ensure_index("state")
        self.tasks.ensure_index("formula_alphabetical")

        # Search index for molecules
        self.molecules.ensure_index("molecule_id")
        self.molecules.ensure_index("last_updated")
        self.molecules.ensure_index("task_ids")
        self.molecules.ensure_index("formula_alphabetical")

        # Search index for thermo
        self.thermo.ensure_index("molecule_id")
        self.thermo.ensure_index("task_id")
        self.thermo.ensure_index("solvent")
        self.thermo.ensure_index("lot_solvent")
        self.thermo.ensure_index("property_id")
        self.thermo.ensure_index("last_updated")
        self.thermo.ensure_index("formula_alphabetical")

        # Search index for redox
        self.redox.ensure_index("molecule_id")
        self.redox.ensure_index("solvent")
        self.redox.ensure_index("lot_solvent")
        self.redox.ensure_index("property_id")
        self.redox.ensure_index("last_updated")
        self.redox.ensure_index("formula_alphabetical")

    def prechunk(self, number_splits: int) -> Iterable[Dict]:  # pragma: no cover
        """Prechunk the builder for distributed computation"""

        temp_query = dict(self.query)
        temp_query["deprecated"] = False

        self.logger.info("Finding documents to process")
        all_mols = list(
            self.molecules.query(
                temp_query, [self.molecules.key, "formula_alphabetical"]
            )
        )

        processed_docs = set([e for e in self.redox.distinct("molecule_id")])
        to_process_docs = {d[self.molecules.key] for d in all_mols} - processed_docs
        to_process_forms = {
            d["formula_alphabetical"]
            for d in all_mols
            if d[self.molecules.key] in to_process_docs
        }

        N = ceil(len(to_process_forms) / number_splits)

        for formula_chunk in grouper(to_process_forms, N):
            yield {"query": {"formula_alphabetical": {"$in": list(formula_chunk)}}}

    def get_items(self) -> Iterator[List[Dict]]:
        """
        Gets all items to process into redox documents.
        This does no datetime checking; relying on on whether
        task_ids are included in the orbitals Store

        Returns:
            generator or list relevant tasks and molecules to process into documents
        """

        self.logger.info("Redox builder started")
        self.logger.info("Setting indexes")
        self.ensure_indexes()

        # Save timestamp to mark buildtime
        self.timestamp = datetime.utcnow()

        # Get all processed molecules
        temp_query = dict(self.query)
        temp_query["deprecated"] = False

        self.logger.info("Finding documents to process")
        all_mols = list(
            self.molecules.query(
                temp_query, [self.molecules.key, "formula_alphabetical"]
            )
        )

        processed_docs = set([e for e in self.redox.distinct("molecule_id")])
        to_process_docs = {d[self.molecules.key] for d in all_mols} - processed_docs
        to_process_forms = {
            d["formula_alphabetical"]
            for d in all_mols
            if d[self.molecules.key] in to_process_docs
        }

        self.logger.info(f"Found {len(to_process_docs)} unprocessed documents")
        self.logger.info(f"Found {len(to_process_forms)} unprocessed formulas")

        # Set total for builder bars to have a total
        self.total = len(to_process_forms)

        for formula in to_process_forms:
            mol_query = dict(temp_query)
            mol_query["formula_alphabetical"] = formula
            molecules = list(self.molecules.query(criteria=mol_query))

            yield molecules

    def process_item(self, items: List[Dict]) -> List[Dict]:
        """
        Process the tasks into a RedoxDoc

        Args:
            tasks List[Dict] : a list of MoleculeDocs in dict form

        Returns:
            [dict] : a list of new redox docs
        """

        mols = [MoleculeDoc(**item) for item in items]
        formula = mols[0].formula_alphabetical
        mol_ids = [m.molecule_id for m in mols]
        self.logger.debug(f"Processing {formula} : {mol_ids}")

        redox_docs = list()

        # Group by (covalent) molecular graph connectivity
        group_by_graph = self._group_by_graph(mols)

        for graph_group in group_by_graph.values():
            # Molecule docs will be grouped by charge
            charges: Dict[int, Any] = dict()

            for gg in graph_group:
                # First, grab relevant MoleculeThermoDocs and identify possible IE/EA single-points
                thermo_docs = [
                    MoleculeThermoDoc(**e)
                    for e in self.thermo.query({"molecule_id": gg.molecule_id})
                ]

                if len(thermo_docs) == 0:
                    # Current building scheme requires a MoleculeThermoDoc
                    continue

                ie_sp_task_ids = [
                    e["task_id"]
                    for e in gg.entries
                    if e["charge"] == gg.charge + 1
                    and e["task_type"] == "Single Point"
                    and e["output"].get("final_energy")
                ]
                ie_tasks = list()
                for i in ie_sp_task_ids:
                    tdoc = self.tasks.query_one(
                        {
                            "task_id": i,
                            "formula_alphabetical": formula,
                            "orig": {"$exists": True},
                        }
                    )

                    if tdoc is None:
                        try:
                            tdoc = self.tasks.query_one(
                                {
                                    "task_id": int(i),
                                    "formula_alphabetical": formula,
                                    "orig": {"$exists": True},
                                }
                            )
                        except ValueError:
                            tdoc = None

                    if tdoc is None:
                        continue

                    ie_tasks.append(TaskDocument(**tdoc))

                ea_sp_task_ids = [
                    e["task_id"]
                    for e in gg.entries
                    if e["charge"] == gg.charge - 1
                    and e["task_type"] == "Single Point"
                    and e["output"].get("final_energy")
                ]
                ea_tasks = list()
                for i in ea_sp_task_ids:
                    tdoc = self.tasks.query_one(
                        {
                            "task_id": i,
                            "formula_alphabetical": formula,
                            "orig": {"$exists": True},
                        }
                    )

                    if tdoc is None:
                        try:
                            tdoc = self.tasks.query_one(
                                {
                                    "task_id": int(i),
                                    "formula_alphabetical": formula,
                                    "orig": {"$exists": True},
                                }
                            )
                        except ValueError:
                            tdoc = None

                    if tdoc is None:
                        continue

                    ea_tasks.append(TaskDocument(**tdoc))

                grouped_docs = self._collect_by_lot_solvent(
                    thermo_docs, ie_tasks, ea_tasks
                )
                if gg.charge in charges:
                    charges[gg.charge].append((gg, grouped_docs))
                else:
                    charges[gg.charge] = [(gg, grouped_docs)]

            for charge, collection in charges.items():
                for mol, docs in collection:
                    # Get all possible molecules for adiabatic oxidation and reduction
                    red_coll = charges.get(charge - 1, list())
                    ox_coll = charges.get(charge + 1, list())

                    for lot_solv, docset in docs.items():
                        # Collect other molecules that have MoleculeThermoDocs at the
                        # exact same level of theory

                        combined = docset["thermo_doc"].combined_lot_solvent

                        relevant_red = list()
                        relevant_ox = list()

                        for rmol, rdocs in red_coll:
                            if lot_solv in rdocs:
                                if (
                                    rdocs[lot_solv]["thermo_doc"].combined_lot_solvent
                                    == combined
                                ):
                                    relevant_red.append(rdocs[lot_solv])

                        for omol, odocs in ox_coll:
                            if lot_solv in odocs:
                                if (
                                    odocs[lot_solv]["thermo_doc"].combined_lot_solvent
                                    == combined
                                ):
                                    relevant_ox.append(odocs[lot_solv])

                        # Take best options (based on electronic energy), where available
                        if len(relevant_red) == 0:
                            red_doc = None
                        else:
                            red_doc = sorted(
                                relevant_red,
                                key=lambda x: x["thermo_doc"].electronic_energy,
                            )[0]["thermo_doc"]

                        if len(relevant_ox) == 0:
                            ox_doc = None
                        else:
                            ox_doc = sorted(
                                relevant_ox,
                                key=lambda x: x["thermo_doc"].electronic_energy,
                            )[0]["thermo_doc"]

                        ea_doc = docset.get("ea_doc")
                        ie_doc = docset.get("ie_doc")

                        redox_docs.append(
                            RedoxDoc.from_docs(
                                base_molecule_doc=mol,
                                base_thermo_doc=docset["thermo_doc"],
                                red_doc=red_doc,
                                ox_doc=ox_doc,
                                ea_doc=ea_doc,
                                ie_doc=ie_doc,
                            )
                        )

        self.logger.debug(f"Produced {len(redox_docs)} redox docs for {formula}")

        return jsanitize(
            [doc.model_dump() for doc in redox_docs if doc is not None], allow_bson=True
        )

    def update_targets(self, items: List[List[Dict]]):
        """
        Inserts the new documents into the orbitals collection

        Args:
            items [[dict]]: A list of documents to update
        """

        docs = list(chain.from_iterable(items))  # type: ignore

        # Add timestamp
        for item in docs:
            item.update(
                {
                    "_bt": self.timestamp,
                }
            )

        molecule_ids = list({item["molecule_id"] for item in docs})

        if len(items) > 0:
            self.logger.info(f"Updating {len(docs)} redox documents")
            self.redox.remove_docs({self.redox.key: {"$in": molecule_ids}})
            self.redox.update(
                docs=docs,
                key=["molecule_id", "solvent"],
            )
        else:
            self.logger.info("No items to update")

    @staticmethod
    def _group_by_graph(mol_docs: List[MoleculeDoc]) -> Dict[int, List[MoleculeDoc]]:
        """
        Group molecule docs by molecular graph connectivity

        :param entries: List of entries (dicts derived from TaskDocuments)
        :return: Grouped molecule entries
        """

        graph_hashes_nometal: List[str] = list()
        results = defaultdict(list)

        # Within each group, group by the covalent molecular graph
        for t in mol_docs:
            mol = confirm_molecule(t.molecule)

            mol_nometal = copy.deepcopy(mol)

            if mol.composition.alphabetical_formula not in [m + "1" for m in metals]:
                mol_nometal.remove_species(metals)

            mol_nometal.set_charge_and_spin(0)
            gh_nometal = get_graph_hash(mol_nometal, node_attr="specie")

            match = None
            for i, gh in enumerate(graph_hashes_nometal):
                if gh_nometal == gh:
                    match = i
                    break

            if match is None:
                results[len(graph_hashes_nometal)].append(t)
                graph_hashes_nometal.append(gh_nometal)
            else:
                results[match].append(t)

        return results

    @staticmethod
    def _collect_by_lot_solvent(
        thermo_docs: List[MoleculeThermoDoc],
        ie_docs: List[TaskDocument],
        ea_docs: List[TaskDocument],
    ) -> Dict[str, Any]:
        """
        For a given MoleculeDoc, group potential MoleculeThermoDocs and TaskDocs for
        IE/EA calculations based on level of theory and solvent.

        Args:
            thermo_docs (list of MoleculeThermoDocs): List of MoleculeThermoDocs for this MoleculeDoc
            ie_docs (list of TaskDocuments): List of TaskDocs which could be used
                to calculate vertical ionization energies for this MoleculeDoc
            ea_docs (list of TaskDocuments): List of TaskDocs which could be used
                to calculate vertical electron affinities for this MoleculeDoc:

        Returns:
            dict {<lot_solvent>: {
                        "thermo_doc": MoleculeThermoDoc, "ie_doc": TaskDocument, "ea_doc": TaskDocument
                    }
                 }
        """

        def _lot_solv(doc: Union[MoleculeThermoDoc, TaskDocument]):
            if isinstance(doc, MoleculeThermoDoc):
                if doc.correction:
                    return doc.correction_lot_solvent
            return doc.lot_solvent

        thermo_grouped = groupby(sorted(thermo_docs, key=_lot_solv), key=_lot_solv)
        ie_grouped = groupby(sorted(ie_docs, key=_lot_solv), key=_lot_solv)
        ea_grouped = groupby(sorted(ea_docs, key=_lot_solv), key=_lot_solv)

        groups = dict()

        for k, g in thermo_grouped:
            g_list = list(g)

            # Should never be more than one MoleculeThermoDoc per MoleculeDoc
            # Just for safety...
            if len(g_list) > 1:
                g_list_sorted = sorted(g_list, key=lambda x: x.electronic_energy)
                this_thermo_doc = g_list_sorted[0]
            else:
                this_thermo_doc = g_list[0]

            groups[k] = {"thermo_doc": this_thermo_doc}

        for k, g in ie_grouped:
            # Must be a MoleculeThermoDoc to make a RedoxDoc
            if k not in groups:
                continue

            this_ie_doc = sorted(list(g), key=lambda x: x.output.final_energy)[0]
            groups[k]["ie_doc"] = this_ie_doc

        for k, g in ea_grouped:
            # Must be a MoleculeThermoDoc to make a RedoxDoc
            if k not in groups:
                continue

            this_ea_doc = sorted(list(g), key=lambda x: x.output.final_energy)[0]
            groups[k]["ea_doc"] = this_ea_doc

        return groups



import traceback
from datetime import datetime
from itertools import chain
from typing import Dict, List

from maggma.builders import GroupBuilder
from maggma.core import Store

from emmet.core.feff.task import TaskDocument as FEFFTaskDocument
from emmet.core.utils import jsanitize
from emmet.core.xas import XASDoc


class XASBuilder(GroupBuilder):
    """
    Generates XAS Docs from FEFF tasks

    # TODO: Generate MPID from materials collection rather than from task metadata
    """

    def __init__(self, tasks: Store, xas: Store, num_samples: int = 200, **kwargs):
        self.tasks = tasks
        self.xas = xas
        self.num_samples = num_samples
        self.kwargs = kwargs

        super().__init__(source=tasks, target=xas, grouping_keys=["mp_id"])
        self._target_keys_field = "xas_ids"

    def process_item(self, spectra: List[Dict]) -> Dict:
        # TODO: Change this to do structure matching against materials collection
        mpid = spectra[0]["mp_id"]

        self.logger.debug(f"Processing: {mpid}")

        tasks = [FEFFTaskDocument(**task) for task in spectra]

        try:
            docs = XASDoc.from_task_docs(tasks, material_id=mpid)
            processed = [d.model_dump() for d in docs]

            for d in processed:
                d.update({"state": "successful"})
        except Exception as e:
            self.logger.error(traceback.format_exc())
            processed = [
                {
                    "error": str(e),
                    "state": "failed",
                    "task_ids": list(d.task_id for d in tasks),
                }
            ]

        update_doc = {
            "_bt": datetime.utcnow(),
        }
        for d in processed:
            d.update({k: v for k, v in update_doc.items() if k not in d})

        return jsanitize(processed, allow_bson=True)

    def update_targets(self, items):
        """
        Group buidler isn't designed for many-to-many so we unwrap that here
        """

        items = list(filter(None.__ne__, chain.from_iterable(items)))
        super().update_targets(items)



import operator
from datetime import datetime
from functools import lru_cache
from itertools import chain
from math import ceil
from typing import Any, Iterator, Dict, List, Optional
from collections import defaultdict

from maggma.builders import Builder
from maggma.stores import MongoStore
from maggma.utils import grouper
from pymatgen.entries.computed_entries import ComputedEntry, ComputedStructureEntry
from pymatgen.entries.compatibility import MaterialsProject2020Compatibility
from pymatgen.analysis.phase_diagram import PhaseDiagram, Composition

from emmet.core.electrode import InsertionElectrodeDoc, ConversionElectrodeDoc
from emmet.core.structure_group import StructureGroupDoc, _get_id_num
from emmet.core.utils import jsanitize
from emmet.builders.settings import EmmetBuildSettings


def s_hash(el):
    return el.data["comp_delith"]


REDOX_ELEMENTS = [
    "Ti",
    "V",
    "Cr",
    "Mn",
    "Fe",
    "Co",
    "Ni",
    "Cu",
    "Nb",
    "Mo",
    "Ag",
    "Sn",
    "Sb",
    "W",
    "Re",
    "Bi",
    "C",
]

WORKING_IONS = ["Li", "Na", "K", "Mg", "Ca", "Zn", "Al"]

MAT_PROPS = ["structure", "material_id", "formula_pretty", "entries"]

sg_fields = ["number", "hall_number", "international", "hall", "choice"]


def generic_groupby(list_in, comp=operator.eq):
    """
    Group a list of unsortable objects
    Args:
        list_in: A list of generic objects
        comp: (Default value = operator.eq) The comparator
    Returns:
        [int] list of labels for the input list
    """
    list_out = [None] * len(list_in)
    label_num = 0
    for i1, ls1 in enumerate(list_out):
        if ls1 is not None:
            continue
        list_out[i1] = label_num
        for i2, ls2 in list(enumerate(list_out))[i1 + 1 :]:
            if comp(list_in[i1], list_in[i2]):
                if list_out[i2] is None:
                    list_out[i2] = list_out[i1]
                else:
                    list_out[i1] = list_out[i2]
                    label_num -= 1
        label_num += 1
    return list_out


default_build_settings = EmmetBuildSettings()


class StructureGroupBuilder(Builder):
    def __init__(
        self,
        materials: MongoStore,
        sgroups: MongoStore,
        working_ion: str,
        query: Optional[dict] = None,
        ltol: float = default_build_settings.LTOL,
        stol: float = default_build_settings.STOL,
        angle_tol: float = default_build_settings.ANGLE_TOL,
        check_newer: bool = True,
        chunk_size: int = 1000,
        **kwargs,
    ):
        """
        Aggregate materials entries into sgroups that are topotactically similar to each other.
        This is an incremental builder that makes ensures that each materials id belongs to one StructureGroupDoc
        document
        Args:
            materials (Store): Store of materials documents that contains the structures
            sgroups (Store): Store of grouped material ids
            query (dict): dictionary to limit materials to be analyzed ---
                            only applied to the materials when we need to group structures
                            the phase diagram is still constructed with the entire set
            chunk_size (int): Size of chemsys chunks to process at any one time.
        """
        self.materials = materials
        self.sgroups = sgroups
        self.working_ion = working_ion
        self.query = query if query else {}
        self.ltol = ltol
        self.stol = stol
        self.angle_tol = angle_tol
        self.check_newer = check_newer
        self.chunk_size = chunk_size

        self.query[
            "deprecated"
        ] = False  # Ensure only non-deprecated materials are chosen

        super().__init__(
            sources=[materials], targets=[sgroups], chunk_size=chunk_size, **kwargs
        )

    def prechunk(self, number_splits: int) -> Iterator[Dict]:  # pragma: no cover
        """
        Prechunk method to perform chunking by the key field
        """
        q = dict(self.query)

        all_chemsys = self.materials.distinct("chemsys", criteria=q)

        new_chemsys_list = []

        for chemsys in all_chemsys:
            elements = [
                element for element in chemsys.split("-") if element != self.working_ion
            ]
            new_chemsys = "-".join(sorted(elements))
            new_chemsys_list.append(new_chemsys)

        N = ceil(len(new_chemsys_list) / number_splits)

        for split in grouper(new_chemsys_list, N):
            new_split_add = []
            for chemsys in split:
                elements = [element for element in chemsys.split("-")] + [
                    self.working_ion
                ]
                new_chemsys = "-".join(sorted(elements))
                new_split_add.append(new_chemsys)

            yield {"query": {"chemsys": {"$in": new_split_add + split}}}

    def get_items(self):
        """
        Summary of the steps:
        - query the materials database for different chemical systems that satisfies the base query
          "contains redox element and working ion"
        - Get the full chemsys list of interest
        - The main loop is over all these chemsys.  within the main loop:
            - get newest timestamp for the material documents (max_mat_time)
            - get the oldest timestamp for the target documents (min_target_time)
            - if min_target_time is < max_mat_time then nuke all the target documents
        """
        # All potentially interesting chemsys must contain the working ion
        base_query = {
            "$and": [
                self.query.copy(),
                {"elements": {"$in": REDOX_ELEMENTS}},
                {"elements": {"$in": [self.working_ion]}},
            ]
        }
        self.logger.debug(f"Initial Chemsys QUERY: {base_query}")

        # get a chemsys that only contains the working ion since the working ion
        # must be present for there to be voltage steps
        all_chemsys = self.materials.distinct("chemsys", criteria=base_query)
        # Contains the working ion but not ONLY the working ion
        all_chemsys = [
            *filter(
                lambda x: self.working_ion in x and len(x) > 1,
                [chemsys_.split("-") for chemsys_ in all_chemsys],
            )
        ]

        self.logger.debug(
            f"Performing initial checks on {len(all_chemsys)} chemical systems containing redox elements w/ or w/o wion"
        )
        self.total = len(all_chemsys)

        for chemsys_l in all_chemsys:
            chemsys = "-".join(sorted(chemsys_l))
            chemsys_wo = "-".join(sorted(set(chemsys_l) - {self.working_ion}))
            chemsys_query = {
                "$and": [{"chemsys": {"$in": [chemsys_wo, chemsys]}}, self.query.copy()]
            }
            self.logger.debug(f"QUERY: {chemsys_query}")
            all_mats_in_chemsys = list(
                self.materials.query(
                    criteria=chemsys_query,
                    properties=MAT_PROPS + [self.materials.last_updated_field],
                )
            )
            self.logger.debug(
                f"Found {len(all_mats_in_chemsys)} materials in {chemsys_wo}"
            )
            if self.check_newer:
                all_target_docs = list(
                    self.sgroups.query(
                        criteria={"chemsys": chemsys},
                        properties=[
                            "group_id",
                            self.sgroups.last_updated_field,
                            "material_ids",
                        ],
                    )
                )
                self.logger.debug(
                    f"Found {len(all_target_docs)} Grouped documents in {chemsys_wo}"
                )

                mat_times = [
                    mat_doc[self.materials.last_updated_field]
                    for mat_doc in all_mats_in_chemsys
                ]
                max_mat_time = max(mat_times, default=datetime.min)
                self.logger.debug(
                    f"The newest material doc was generated at {max_mat_time}."
                )

                target_times = [
                    g_doc[self.materials.last_updated_field]
                    for g_doc in all_target_docs
                ]
                min_target_time = min(target_times, default=datetime.max)
                self.logger.debug(
                    f"The newest GROUP doc was generated at {min_target_time}."
                )

                mat_ids = set(
                    [mat_doc["material_id"] for mat_doc in all_mats_in_chemsys]
                )

                # If any material id is missing or if any material id has been updated
                target_ids = set()
                for g_doc in all_target_docs:
                    target_ids |= set(g_doc["material_ids"])

                self.logger.debug(
                    f"There are {len(mat_ids)} material ids in source database vs {len(target_ids)} in target database."
                )
                if mat_ids == target_ids and max_mat_time < min_target_time:
                    self.logger.info(f"Skipping chemsys {chemsys}.")
                    yield None
                elif len(target_ids) == 0:
                    self.logger.info(
                        f"No documents in chemsys {chemsys} in the target database."
                    )
                    yield {"chemsys": chemsys, "materials": all_mats_in_chemsys}
                else:
                    self.logger.info(
                        f"Nuking all {len(target_ids)} documents in chemsys {chemsys} in the target database."
                    )
                    self._remove_targets(list(target_ids))
                    yield {"chemsys": chemsys, "materials": all_mats_in_chemsys}
            else:
                yield {"chemsys": chemsys, "materials": all_mats_in_chemsys}

    def update_targets(self, items: List):
        items = list(filter(None, chain.from_iterable(items)))
        if len(items) > 0:
            self.logger.info("Updating {} sgroups documents".format(len(items)))
            for struct_group_dict in items:
                struct_group_dict[self.sgroups.last_updated_field] = datetime.utcnow()
            self.sgroups.update(docs=items, key=["group_id"])
        else:
            self.logger.info("No items to update")

    def _entry_from_mat_doc(self, mdoc):
        # Note since we are just structure grouping we don't need to be careful with energy or correction
        # All of the energy analysis is left to other builders
        entries = [
            ComputedStructureEntry.from_dict(v) for v in mdoc["entries"].values()
        ]
        if len(entries) == 1:
            return entries[0]
        else:
            if "GGA+U" in mdoc["entries"].keys():
                return ComputedStructureEntry.from_dict(mdoc["entries"]["GGA+U"])
            elif "GGA" in mdoc["entries"].keys():
                return ComputedStructureEntry.from_dict(mdoc["entries"]["GGA"])
            else:
                return None

    def process_item(self, item: Any) -> Any:
        if item is None:
            return None
        entries = [*map(self._entry_from_mat_doc, item["materials"])]
        compatibility = MaterialsProject2020Compatibility()
        processed_entries = compatibility.process_entries(entries=entries)
        s_groups = StructureGroupDoc.from_ungrouped_structure_entries(
            entries=processed_entries,
            ignored_specie=self.working_ion,
            ltol=self.ltol,
            stol=self.stol,
            angle_tol=self.angle_tol,
        )
        return [sg.model_dump() for sg in s_groups]

    def _remove_targets(self, rm_ids):
        self.sgroups.remove_docs({"material_ids": {"$in": rm_ids}})


class InsertionElectrodeBuilder(Builder):
    def __init__(
        self,
        grouped_materials: MongoStore,
        thermo: MongoStore,
        insertion_electrode: MongoStore,
        query: Optional[Dict] = None,
        strip_structures: bool = False,
        **kwargs,
    ):
        self.grouped_materials = grouped_materials
        self.insertion_electrode = insertion_electrode
        self.thermo = thermo
        self.query = query if query else {}
        self.strip_structures = strip_structures

        super().__init__(
            sources=[self.grouped_materials, self.thermo],
            targets=[self.insertion_electrode],
            **kwargs,
        )

    def prechunk(self, number_splits: int) -> Iterator[Dict]:
        """
        Prechunk method to perform chunking by the key field
        """
        q = dict(self.query)

        keys = self.grouped_materials.distinct(self.grouped_materials.key, criteria=q)

        N = ceil(len(keys) / number_splits)
        for split in grouper(keys, N):
            yield {"query": {self.grouped_materials.key: {"$in": list(split)}}}

    def get_items(self):
        """
        Get items
        """

        @lru_cache(1000)
        def get_working_ion_entry(working_ion):
            with self.thermo as store:
                working_ion_docs = [*store.query({"chemsys": working_ion})]
            best_wion = min(working_ion_docs, key=lambda x: x["energy_per_atom"])
            return best_wion

        def get_thermo_docs(mat_ids):
            self.logger.debug(
                f"Looking for {len(mat_ids)} material_id in the Thermo DB."
            )
            self.thermo.connect()
            thermo_docs = list(
                self.thermo.query(
                    {"$and": [{"material_id": {"$in": mat_ids}}]},
                    properties=[
                        "material_id",
                        "_sbxn",
                        "thermo",
                        "entries",
                        "energy_type",
                        "energy_above_hull",
                    ],
                )
            )

            self.logger.debug(f"Found for {len(thermo_docs)} Thermo Documents.")
            if len(thermo_docs) != len(mat_ids):
                missing_ids = set(mat_ids) - set(
                    [t_["material_id"] for t_ in thermo_docs]
                )
                self.logger.warn(
                    f"The following ids are missing from the entries in thermo {missing_ids}.\n"
                    "The is likely due to the fact that a calculation other than GGA or GGA+U was "
                    "validated for the materials builder."
                )
                return None

            # if len(item["ignored_species"]) != 1:
            #     raise ValueError(
            #         "Insertion electrode can only be defined for one working ion species"
            #     )

            return thermo_docs
            # return {
            #     "group_id": item["group_id"],
            #     "working_ion_doc": working_ion_doc,
            #     "working_ion": item["ignored_species"][0],
            #     "thermo_docs": thermo_docs,
            # }

        q_ = {"$and": [self.query, {"has_distinct_compositions": True}]}
        self.total = self.grouped_materials.count(q_)
        for group_doc in self.grouped_materials.query(q_):
            working_ion_doc = get_working_ion_entry(group_doc["ignored_specie"])
            thermo_docs = get_thermo_docs(group_doc["material_ids"])
            if thermo_docs:
                yield {
                    "group_id": group_doc["group_id"],
                    "working_ion_doc": working_ion_doc,
                    "working_ion": group_doc["ignored_specie"],
                    "thermo_docs": thermo_docs,
                }
            else:
                yield None

    def process_item(self, item) -> Dict:
        """
        - Add volume information to each entry to create the insertion electrode document
        - Add the host structure
        """
        if item is None:
            return None  # type: ignore
        self.logger.debug(
            f"Working on {item['group_id']} with {len(item['thermo_docs'])}"
        )

        entries = [
            tdoc_["entries"][tdoc_["energy_type"]] for tdoc_ in item["thermo_docs"]
        ]

        entries = list(map(ComputedStructureEntry.from_dict, entries))

        working_ion_entry = ComputedEntry.from_dict(
            item["working_ion_doc"]["entries"][item["working_ion_doc"]["energy_type"]]
        )

        decomp_energies = {
            d_["material_id"]: d_["energy_above_hull"] for d_ in item["thermo_docs"]
        }

        for ient in entries:
            ient.data["volume"] = ient.structure.volume
            ient.data["decomposition_energy"] = decomp_energies[
                ient.data["material_id"]
            ]

        ie = InsertionElectrodeDoc.from_entries(
            grouped_entries=entries,
            working_ion_entry=working_ion_entry,
            battery_id=item["group_id"],
            strip_structures=self.strip_structures,
        )
        if ie is None:
            return None  # type: ignore
            # {"failed_reason": "unable to create InsertionElectrode document"}
        return jsanitize(ie.model_dump())

    def update_targets(self, items: List):
        items = list(filter(None, items))
        if len(items) > 0:
            self.logger.info("Updating {} battery documents".format(len(items)))
            for struct_group_dict in items:
                struct_group_dict[
                    self.grouped_materials.last_updated_field
                ] = datetime.utcnow()
            self.insertion_electrode.update(docs=items, key=["battery_id"])
        else:
            self.logger.info("No items to update")


class ConversionElectrodeBuilder(Builder):
    def __init__(
        self,
        phase_diagram_store: MongoStore,
        conversion_electrode_store: MongoStore,
        working_ion: str,
        thermo_type: str,
        query: Optional[dict] = None,
        **kwargs,
    ):
        self.phase_diagram_store = phase_diagram_store
        self.conversion_electrode_store = conversion_electrode_store
        self.working_ion = working_ion
        self.thermo_type = thermo_type
        self.query = query if query else {}
        self.kwargs = kwargs

        self.phase_diagram_store.key = "phase_diagram_id"
        self.conversion_electrode_store.key = "conversion_electrode_id"

        super().__init__(
            sources=[self.phase_diagram_store],
            targets=[self.conversion_electrode_store],
            **kwargs,
        )

    def prechunk(self, number_splits: int) -> Iterator[Dict]:
        """
        Prechunk method to perform chunking by the key field
        """
        q = dict(self.query)
        keys = self.phase_diagram_store.distinct(
            self.phase_diagram_store.key, criteria=q
        )
        N = ceil(len(keys) / number_splits)
        for split in grouper(keys, N):
            yield {"query": {self.phase_diagram_store.key: {"$in": list(split)}}}

    def get_items(self):
        """
        Get items. Phase diagrams are filtered such that only PDs with chemical systems containing
        the working ion and the specified "thermo_type", or functional, are chosen.
        """

        all_chemsys = self.phase_diagram_store.distinct("chemsys")

        chemsys_w_wion = [c for c in all_chemsys if self.working_ion in c]

        q = {
            "$and": [
                dict(self.query),
                {"thermo_type": self.thermo_type},
                {"chemsys": {"$in": chemsys_w_wion}},
            ]
        }

        for phase_diagram_doc in self.phase_diagram_store.query(criteria=q):
            yield phase_diagram_doc

    def process_item(self, item) -> Dict:
        """
        - For each phase diagram doc, find all possible conversion electrodes and create conversion electrode docs
        """
        # To work around "el_refs" serialization issue (#576)
        _pd = PhaseDiagram.from_dict(item["phase_diagram"])
        _entries = _pd.all_entries
        pd = PhaseDiagram(entries=_entries)

        most_wi = defaultdict(lambda: (-1, None))  # type: dict
        n_elements = pd.dim
        # Only using entries on convex hull for now
        for entry in pd.stable_entries:
            if len(entry.composition.elements) != n_elements:
                continue
            composition_dict = entry.composition.as_dict()
            composition_dict.pop(self.working_ion)
            composition_without_wi = Composition.from_dict(composition_dict)
            red_form, num_form = composition_without_wi.get_reduced_formula_and_factor()
            n_wi = entry.composition.get_el_amt_dict()[self.working_ion]
            most_wi[red_form] = max(
                most_wi[red_form], (n_wi / num_form, entry.composition)
            )

        new_docs = []
        unique_reaction_compositions = set()
        reaction_compositions = []
        for k, v in most_wi.items():
            if v[1] is not None:
                # Get lowest material_id with matching composition
                material_ids = [
                    (
                        lambda x: x.data["material_id"]
                        if x.composition.reduced_formula == v[1].reduced_formula
                        else None
                    )(e)
                    for e in pd.entries
                ]
                material_ids = list(filter(None, material_ids))
                lowest_id = min(material_ids, key=_get_id_num)
                conversion_electrode_doc = (
                    ConversionElectrodeDoc.from_composition_and_pd(
                        comp=v[1],
                        pd=pd,
                        working_ion_symbol=self.working_ion,
                        battery_id=f"{lowest_id}_{self.working_ion}",
                        thermo_type=self.thermo_type,
                    )
                )
                # Get reaction entry_ids
                comps = set()
                for c in conversion_electrode_doc.reaction["reactants"].keys():
                    comps.add(c)
                    unique_reaction_compositions.add(c)
                for c in conversion_electrode_doc.reaction["products"].keys():
                    comps.add(c)
                    unique_reaction_compositions.add(c)
                reaction_compositions.append(comps)
                new_docs.append(jsanitize(conversion_electrode_doc.model_dump()))

        entry_id_mapping = {}
        for c in unique_reaction_compositions:
            relevant_entry_data = []
            for e in pd.entries:
                if e.composition == Composition(c):
                    relevant_entry_data.append((e.energy_per_atom, e.entry_id))
            relevant_entry_data.sort(key=lambda x: x[0])
            entry_id_mapping[c] = relevant_entry_data[0][1]

        for i, comps in enumerate(reaction_compositions):
            mapping = {}
            for c in comps:
                mapping[c] = entry_id_mapping[c]
            new_docs[i]["formula_id_mapping"] = mapping

        return new_docs  # type: ignore

    def update_targets(self, items: List):
        combined_items = []
        for _items in items:
            _items = list(filter(None, _items))
            combined_items.extend(_items)

        if len(combined_items) > 0:
            self.logger.info(
                "Updating {} conversion battery documents".format(len(combined_items))
            )
            self.conversion_electrode_store.update(
                docs=combined_items, key=["battery_id", "thermo_type"]
            )
        else:
            self.logger.info("No items to update")



from typing import TYPE_CHECKING, Union

import pytest
from maggma.stores import MemoryStore
from matcalc.utils import get_universal_calculator
from pymatgen.core import Lattice, Structure

from emmet.builders.materials.ml import MLBuilder

if TYPE_CHECKING:
    from ase.calculators.calculator import Calculator


material_id = "1234"


@pytest.fixture()
def materials_store():
    materials_store = MemoryStore(key="material_id")
    materials_store.connect()
    # Add fake data to materials store
    materials_store.update(
        {
            "material_id": material_id,
            "structure": Structure.from_spacegroup(
                "Pm-3m", Lattice.cubic(4.2), ["Cs", "Cl"], [[0, 0, 0], [0.5, 0.5, 0.5]]
            ).as_dict(),
            "deprecated": False,
        }
    )
    return materials_store


@pytest.mark.skip(reason="Temporary skip. Needs attention.")
@pytest.mark.parametrize("model", [get_universal_calculator("chgnet"), "m3gnet"])
def test_ml_ip_builder(materials_store: MemoryStore, model: Union[str, "Calculator"]):
    ml_store = MemoryStore(key="material_id")

    builder = MLBuilder(materials=materials_store, ml_potential=ml_store, model=model)
    item = materials_store.query_one()

    result_doc = builder.unary_function(item)

    assert result_doc["material_id"] == material_id
    if isinstance(model, str):
        assert result_doc["model_name"] == model
    assert result_doc["energy"] < -6



from pathlib import Path

import pytest
from maggma.stores import JSONStore, MemoryStore
from monty.serialization import dumpfn, loadfn

from emmet.builders.materials.dielectric import DielectricBuilder
from emmet.builders.vasp.materials import MaterialsBuilder


@pytest.fixture(scope="session")
def tasks_store(test_dir):
    return JSONStore(test_dir / "test_si_tasks.json.gz")


@pytest.fixture(scope="session")
def materials_store(tasks_store):
    materials_store = MemoryStore(key="material_id")
    builder = MaterialsBuilder(tasks=tasks_store, materials=materials_store)
    builder.run()
    return materials_store


@pytest.fixture
def dielectric_store():
    return MemoryStore(key="material_id")


def test_dielectric_builder(tasks_store, dielectric_store, materials_store):
    builder = DielectricBuilder(
        tasks=tasks_store, dielectric=dielectric_store, materials=materials_store
    )
    builder.run()

    assert dielectric_store.count() == 1
    assert dielectric_store.count({"deprecated": False}) == 1


def test_serialization(tmpdir):
    builder = DielectricBuilder(MemoryStore(), MemoryStore(), MemoryStore())

    dumpfn(builder.as_dict(), Path(tmpdir) / "test.json")
    loadfn(Path(tmpdir) / "test.json")



from collections import defaultdict
from datetime import datetime
from itertools import chain
from math import ceil
from typing import Optional, Iterable, Iterator, List, Dict

from pymatgen.core.structure import Molecule
from pymatgen.analysis.molecule_matcher import MoleculeMatcher

from maggma.builders import Builder
from maggma.core import Store
from maggma.utils import grouper

from emmet.core.qchem.task import TaskDocument
from emmet.core.qchem.molecule import MoleculeDoc, evaluate_lot
from emmet.core.molecules.thermo import get_free_energy, MoleculeThermoDoc
from emmet.core.qchem.calc_types import TaskType
from emmet.core.utils import jsanitize
from emmet.builders.settings import EmmetBuildSettings


__author__ = "Evan Spotte-Smith"

SETTINGS = EmmetBuildSettings()

single_mol_thermo = {
    "Zn1": {"enthalpy": 1.481, "entropy": 38.384},
    "Xe1": {"enthalpy": 1.481, "entropy": 40.543},
    "Tl1": {"enthalpy": 1.481, "entropy": 41.857},
    "Ti1": {"enthalpy": 1.481, "entropy": 37.524},
    "Te1": {"enthalpy": 1.481, "entropy": 40.498},
    "Sr1": {"enthalpy": 1.481, "entropy": 39.334},
    "Sn1": {"enthalpy": 1.481, "entropy": 40.229},
    "Si1": {"enthalpy": 1.481, "entropy": 35.921},
    "Sb1": {"enthalpy": 1.481, "entropy": 40.284},
    "Se1": {"enthalpy": 1.481, "entropy": 39.05},
    "S1": {"enthalpy": 1.481, "entropy": 36.319},
    "Rn1": {"enthalpy": 1.481, "entropy": 42.095},
    "Pt1": {"enthalpy": 1.481, "entropy": 41.708},
    "Rb1": {"enthalpy": 1.481, "entropy": 39.23},
    "Po1": {"enthalpy": 1.481, "entropy": 41.915},
    "Pb1": {"enthalpy": 1.481, "entropy": 41.901},
    "P1": {"enthalpy": 1.481, "entropy": 36.224},
    "O1": {"enthalpy": 1.481, "entropy": 34.254},
    "Ne1": {"enthalpy": 1.481, "entropy": 34.919},
    "N1": {"enthalpy": 1.481, "entropy": 33.858},
    "Na1": {"enthalpy": 1.481, "entropy": 35.336},
    "Mg1": {"enthalpy": 1.481, "entropy": 35.462},
    "Li1": {"enthalpy": 1.481, "entropy": 31.798},
    "Kr1": {"enthalpy": 1.481, "entropy": 39.191},
    "K1": {"enthalpy": 1.481, "entropy": 36.908},
    "In1": {"enthalpy": 1.481, "entropy": 40.132},
    "I1": {"enthalpy": 1.481, "entropy": 40.428},
    "H1": {"enthalpy": 1.481, "entropy": 26.014},
    "He1": {"enthalpy": 1.481, "entropy": 30.125},
    "Ge1": {"enthalpy": 1.481, "entropy": 38.817},
    "Ga1": {"enthalpy": 1.481, "entropy": 38.609},
    "F1": {"enthalpy": 1.481, "entropy": 34.767},
    "Cu1": {"enthalpy": 1.481, "entropy": 38.337},
    "Cl1": {"enthalpy": 1.481, "entropy": 36.586},
    "Ca1": {"enthalpy": 1.481, "entropy": 36.984},
    "C1": {"enthalpy": 1.481, "entropy": 33.398},
    "Br1": {"enthalpy": 1.481, "entropy": 39.012},
    "Bi1": {"enthalpy": 1.481, "entropy": 41.915},
    "Be1": {"enthalpy": 1.481, "entropy": 32.544},
    "Ba1": {"enthalpy": 1.481, "entropy": 40.676},
    "B1": {"enthalpy": 1.481, "entropy": 33.141},
    "Au1": {"enthalpy": 1.481, "entropy": 41.738},
    "At1": {"enthalpy": 1.481, "entropy": 41.929},
    "As1": {"enthalpy": 1.481, "entropy": 38.857},
    "Ar1": {"enthalpy": 1.481, "entropy": 36.983},
    "Al1": {"enthalpy": 1.481, "entropy": 35.813},
    "Ag1": {"enthalpy": 1.481, "entropy": 39.917},
}


class ThermoBuilder(Builder):
    """
    The ThermoBuilder extracts the highest-quality thermodynamic data from a
    MoleculeDoc (lowest electronic energy, highest level of theory for each
    solvent available).

    This builder constructs MoleculeThermoDocs in two different ways: with and without
    single-point energy corrections.

    Before any documents are constructed, the following steps are taken:
        1. Gather MoleculeDocs by formula
        2. For each doc, identify tasks with thermodynamic information such as
            zero-point energy, enthalpy, and entropy. Collect these "documents
             including complete thermodynamics" (DICTs).
        3. Separately, collect single-point energy calculations (SPECs).
        4. Sort both sets of collected tasks (DICT and SPEC) by solvent

    The first type of doc - those without corrections - can be constructed in
    a straightforward fashion:
        5. For each solvent, grab the best DICT (where "best" is defined as the
            task generated using the highest level of theory with the lowest
            electronic energy)
        6. Convert this TaskDoc to MoleculeThermoDoc

    The second type - those involving single-point energy corrections - are
    generated differently and in a slightly more involved process:
        7. For each of the "best" DICT docs identified in step 5 above:
            7.1 For each solvent, grab the best SPEC
            7.2 Try to match each best SPEC with a matching DICT (meaning that
                the DICT and the SPEC have identical structure) where the DICT
                is calculated at a lower or the same level of theory than the
                SPEC
            7.3 Convert each DICT-SPEC combination to create a MoleculeThermoDoc

    In the case where there are multiple MoleculeThermoDocs made for a given solvent,
    the different MoleculeThermoDocs will be ranked, first by level of theory (for
    a doc made with an energy correction, the scores of the DICT and the SPEC
    levels of theory will be averaged) and then by electronic energy.
    """

    def __init__(
        self,
        tasks: Store,
        molecules: Store,
        thermo: Store,
        query: Optional[Dict] = None,
        settings: Optional[EmmetBuildSettings] = None,
        **kwargs,
    ):
        self.tasks = tasks
        self.molecules = molecules
        self.thermo = thermo
        self.query = query if query else dict()
        self.settings = EmmetBuildSettings.autoload(settings)
        self.kwargs = kwargs

        super().__init__(sources=[tasks, molecules], targets=[thermo], **kwargs)
        # Uncomment in case of issue with mrun not connecting automatically to collections
        # for i in [self.tasks, self.molecules, self.thermo]:
        #     try:
        #         i.connect()
        #     except Exception as e:
        #         print("Could not connect,", e)

    def ensure_indexes(self):
        """
        Ensures indices on the collections needed for building
        """

        # Basic search index for tasks
        self.tasks.ensure_index("task_id")
        self.tasks.ensure_index("last_updated")
        self.tasks.ensure_index("state")
        self.tasks.ensure_index("formula_alphabetical")

        # Search index for molecules
        self.molecules.ensure_index("molecule_id")
        self.molecules.ensure_index("last_updated")
        self.molecules.ensure_index("task_ids")
        self.molecules.ensure_index("formula_alphabetical")

        # Search index for thermo
        self.thermo.ensure_index("molecule_id")
        self.thermo.ensure_index("task_id")
        self.thermo.ensure_index("solvent")
        self.thermo.ensure_index("lot_solvent")
        self.thermo.ensure_index("property_id")
        self.thermo.ensure_index("last_updated")
        self.thermo.ensure_index("formula_alphabetical")

    def prechunk(self, number_splits: int) -> Iterable[Dict]:  # pragma: no cover
        """Prechunk the builder for distributed computation"""

        temp_query = dict(self.query)
        temp_query["deprecated"] = False

        self.logger.info("Finding documents to process")
        all_mols = list(
            self.molecules.query(
                temp_query, [self.molecules.key, "formula_alphabetical"]
            )
        )

        processed_docs = set([e for e in self.thermo.distinct("molecule_id")])
        to_process_docs = {d[self.molecules.key] for d in all_mols} - processed_docs
        to_process_forms = {
            d["formula_alphabetical"]
            for d in all_mols
            if d[self.molecules.key] in to_process_docs
        }

        N = ceil(len(to_process_forms) / number_splits)

        for formula_chunk in grouper(to_process_forms, N):
            yield {"query": {"formula_alphabetical": {"$in": list(formula_chunk)}}}

    def get_items(self) -> Iterator[List[Dict]]:
        """
        Gets all items to process into thermo documents.
        This does no datetime checking; relying on on whether
        task_ids are included in the thermo Store

        Returns:
            generator or list relevant tasks and molecules to process into documents
        """

        self.logger.info("Thermo builder started")
        self.logger.info("Setting indexes")
        self.ensure_indexes()

        # Save timestamp to mark buildtime
        self.timestamp = datetime.utcnow()

        # Get all processed molecules
        temp_query = dict(self.query)
        temp_query["deprecated"] = False

        self.logger.info("Finding documents to process")
        all_mols = list(
            self.molecules.query(
                temp_query, [self.molecules.key, "formula_alphabetical"]
            )
        )

        processed_docs = set([e for e in self.thermo.distinct("molecule_id")])
        to_process_docs = {d[self.molecules.key] for d in all_mols} - processed_docs
        to_process_forms = {
            d["formula_alphabetical"]
            for d in all_mols
            if d[self.molecules.key] in to_process_docs
        }

        self.logger.info(f"Found {len(to_process_docs)} unprocessed documents")
        self.logger.info(f"Found {len(to_process_forms)} unprocessed formulas")

        # Set total for builder bars to have a total
        self.total = len(to_process_forms)

        for formula in to_process_forms:
            mol_query = dict(temp_query)
            mol_query["formula_alphabetical"] = formula
            molecules = list(self.molecules.query(criteria=mol_query))

            yield molecules

    def process_item(self, items: List[Dict]) -> List[Dict]:
        """
        Process the tasks into a MoleculeThermoDoc

        Args:
            items List[dict] : a list of MoleculeDocs in dict form

        Returns:
            [dict] : a list of new thermo docs
        """

        def _add_single_atom_enthalpy_entropy(
            task: TaskDocument, doc: MoleculeThermoDoc
        ):
            initial_mol = task.output.initial_molecule
            # If single atom, try to add enthalpy and entropy
            if len(initial_mol) == 1:
                if doc.total_enthalpy is None or doc.total_entropy is None:
                    formula = initial_mol.composition.alphabetical_formula
                    if formula in single_mol_thermo:
                        vals = single_mol_thermo[formula]
                        doc.total_enthalpy = vals["enthalpy"] * 0.043363
                        doc.total_entropy = vals["entropy"] * 0.000043363
                        doc.translational_enthalpy = vals["enthalpy"] * 0.043363
                        doc.translational_entropy = vals["entropy"] * 0.000043363
                        doc.free_energy = get_free_energy(
                            doc.electronic_energy,
                            vals["enthalpy"],
                            vals["entropy"],
                            convert_energy=False,
                        )
            return doc

        mols = [MoleculeDoc(**item) for item in items]
        formula = mols[0].formula_alphabetical
        mol_ids = [m.molecule_id for m in mols]
        self.logger.debug(f"Processing {formula} : {mol_ids}")

        thermo_docs = list()

        mm = MoleculeMatcher(tolerance=0.000001)

        for mol in mols:
            this_thermo_docs = list()
            # Collect DICTs and SPECs
            thermo_entries = [
                e
                for e in mol.entries
                if e["output"]["enthalpy"] is not None
                and e["output"]["entropy"] is not None
                and e["charge"] == mol.charge
                and e["spin_multiplicity"] == mol.spin_multiplicity
            ]

            sp_entries = list()
            for entry in mol.entries:
                if isinstance(entry["task_type"], TaskType):
                    task_type = entry["task_type"].value
                else:
                    task_type = entry["task_type"]

                if (
                    task_type in ["Single Point", "Force"]
                    and entry["charge"] == mol.charge
                    and entry["spin_multiplicity"] == mol.spin_multiplicity
                ):
                    sp_entries.append(entry)

            # Group both DICTs and SPECs by solvent environment
            by_solvent_dict = defaultdict(list)
            by_solvent_spec = defaultdict(list)
            for entry in thermo_entries:
                by_solvent_dict[entry["solvent"]].append(entry)
            for entry in sp_entries:
                by_solvent_spec[entry["solvent"]].append(entry)

            if len(thermo_entries) == 0:
                without_corrections = by_solvent_spec
            else:
                without_corrections = by_solvent_dict

            # Construct without corrections
            for solvent, entries in without_corrections.items():
                best = sorted(
                    entries,
                    key=lambda x: (
                        sum(evaluate_lot(x["level_of_theory"])),
                        x["energy"],
                    ),
                )[0]
                task = best["task_id"]

                tdoc = self.tasks.query_one(
                    {
                        "task_id": task,
                        "formula_alphabetical": formula,
                        "orig": {"$exists": True},
                    }
                )

                if tdoc is None:
                    try:
                        tdoc = self.tasks.query_one(
                            {
                                "task_id": int(task),
                                "formula_alphabetical": formula,
                                "orig": {"$exists": True},
                            }
                        )
                    except ValueError:
                        tdoc = None

                if tdoc is None:
                    continue

                task_doc = TaskDocument(**tdoc)

                if task_doc is None:
                    continue

                thermo_doc = MoleculeThermoDoc.from_task(
                    task_doc, molecule_id=mol.molecule_id, deprecated=False
                )
                thermo_doc = _add_single_atom_enthalpy_entropy(task_doc, thermo_doc)
                this_thermo_docs.append(thermo_doc)

            # Construct with corrections
            for solvent, entries in by_solvent_spec.items():
                spec_sorted = sorted(
                    entries,
                    key=lambda x: (
                        sum(evaluate_lot(x["level_of_theory"])),
                        x["energy"],
                    ),
                )

                for best_spec in spec_sorted:
                    task_spec = best_spec["task_id"]

                    matching_structures = list()
                    for entry in thermo_entries:
                        mol1 = Molecule.from_dict(entry["molecule"])
                        mol2 = Molecule.from_dict(best_spec["molecule"])
                        if (mm.fit(mol1, mol2) or mol1 == mol2) and (
                            sum(evaluate_lot(best_spec["level_of_theory"]))
                            < sum(evaluate_lot(entry["level_of_theory"]))
                        ):
                            matching_structures.append(entry)

                    if len(matching_structures) == 0:
                        continue

                    best_dict = sorted(
                        matching_structures,
                        key=lambda x: (
                            sum(evaluate_lot(x["level_of_theory"])),
                            x["energy"],
                        ),
                    )[0]
                    task_dict = best_dict["task_id"]

                    tdict = self.tasks.query_one({"task_id": task_dict})
                    if tdict is None:
                        try:
                            tdict = self.tasks.query_one({"task_id": int(task_dict)})
                        except ValueError:
                            tdict = None

                    tspec = self.tasks.query_one({"task_id": task_spec})
                    if tspec is None:
                        try:
                            tspec = self.tasks.query_one({"task_id": int(task_spec)})
                        except ValueError:
                            tspec = None

                    if tdict is None or tspec is None:
                        continue

                    task_doc_dict = TaskDocument(**tdict)
                    task_doc_spec = TaskDocument(**tspec)
                    thermo_doc = MoleculeThermoDoc.from_task(
                        task_doc_dict,
                        correction_task=task_doc_spec,
                        molecule_id=mol.molecule_id,
                        deprecated=False,
                    )
                    thermo_doc = _add_single_atom_enthalpy_entropy(
                        task_doc_dict, thermo_doc
                    )
                    this_thermo_docs.append(thermo_doc)
                    break

            docs_by_solvent = defaultdict(list)
            for doc in this_thermo_docs:
                if doc.correction_solvent is not None:
                    docs_by_solvent[doc.correction_solvent].append(doc)
                else:
                    docs_by_solvent[doc.solvent].append(doc)

            # If multiple documents exist for the same solvent, grab the best one
            for _, collection in docs_by_solvent.items():
                with_eval_e = list()
                for member in collection:
                    if member.correction_level_of_theory is None:
                        with_eval_e.append(
                            (
                                member,
                                sum(evaluate_lot(member.level_of_theory)),
                                member.electronic_energy,
                            )
                        )
                    else:
                        dict_lot = sum(evaluate_lot(member.level_of_theory))
                        spec_lot = sum(evaluate_lot(member.correction_level_of_theory))
                        with_eval_e.append(
                            (
                                member,
                                (dict_lot + spec_lot) / 2,
                                member.electronic_energy,
                            )
                        )

                thermo_docs.append(
                    sorted(with_eval_e, key=lambda x: (x[1], x[2]))[0][0]
                )

        self.logger.debug(f"Produced {len(thermo_docs)} thermo docs for {formula}")

        return jsanitize([doc.model_dump() for doc in thermo_docs], allow_bson=True)

    def update_targets(self, items: List[List[Dict]]):
        """
        Inserts the new thermo docs into the thermo collection

        Args:
            items [[dict]]: A list of documents to update
        """

        docs = list(chain.from_iterable(items))  # type: ignore

        # Add timestamp
        for item in docs:
            item.update(
                {
                    "_bt": self.timestamp,
                }
            )

        molecule_ids = list({item["molecule_id"] for item in docs})

        if len(items) > 0:
            self.logger.info(f"Updating {len(docs)} thermo documents")
            self.thermo.remove_docs({self.thermo.key: {"$in": molecule_ids}})
            self.thermo.update(
                docs=docs,
                key=["molecule_id", "solvent"],
            )
        else:
            self.logger.info("No items to update")



from emmet.core.base import EmmetMeta
import pytest
from maggma.stores import JSONStore, MemoryStore

from emmet.builders.materials.chemenv import ChemEnvBuilder
from emmet.builders.materials.oxidation_states import OxidationStatesBuilder


@pytest.fixture(scope="session")
def fake_materials(test_dir):
    entries = JSONStore(test_dir / "LiTiO2_batt.json", key="entry_id")
    entries.connect()

    materials_store = MemoryStore(key="material_id")
    materials_store.connect()

    for doc in entries.query():
        builder_meta = EmmetMeta(license="BY-C").model_dump()
        materials_store.update(
            {
                "material_id": doc["entry_id"],
                "structure": doc["structure"],
                "deprecated": False,
                "builder_meta": builder_meta,
            }
        )
    return materials_store


def test_chemenvstore(fake_materials):
    oxi_store = MemoryStore()
    builder = OxidationStatesBuilder(
        materials=fake_materials, oxidation_states=oxi_store
    )
    builder.run()
    chemenv_store = MemoryStore()
    builder2 = ChemEnvBuilder(oxidation_states=oxi_store, chemenv=chemenv_store)
    builder2.run()
    assert chemenv_store.count() == 6
    assert all([isinstance(d["composition"], dict) for d in chemenv_store.query()])
